const { calculateObjectSize } = require("bson");
const { isObjectIdOrHexString } = require("mongoose");

exports.GameManager = class GameManager{//Database, contains games, retrieval functions, and base checks
    constructor(){
        this.gamesInWait = {};//Format : gameName: {gameToPlay, playerName}
        this.gamesInPlay = {};//Format : gameName: {gameType, playerHost, playerJoin, isHostTurn, gameData};
    }
    //Gets for Games in Wait
    getGameNames(){//gets the names/keys of the games in wait
        var getGames = this.gamesInWait;
        return Object.keys(getGames);
    }
    getGameValues(){//gets the values/data of the games in wait.
        var getValues = this.gamesInWait;
        return Object.values(getValues);
    }
    getGameKeyValues(aGameName){//gets the values/data of a specific game in games in wait
        var gameValues = this.gamesInWait[aGameName];
        return Object.values(gameValues);
    }
    getPlayerName(aGameName){//gets the name of the waiting host player in games in wait
        var aPlayerName = this.gamesInWait[aGameName];
        return aPlayerName.playerName;
    }
    getGameDataWithKey(aGameName){//returns a game in games in wait with the key added to the array
        console.log("getGameDataWithKey Called")
        var gameValues = this.gamesInWait[aGameName];
        var holdValues = Object.values(gameValues);
        holdValues.unshift(aGameName);
        return holdValues;
    }

    //Gets for Games in Play
    getGameInPlayNames(){//gets the list of games in play.
        var getGames = this.gamesInPlay;
        return Object.keys(getGames);
    }
    getGameInPlayJoinPlayerName(aGameName){//gets the name of the joining player from a game in play
        var aJoinName = this.gamesInPlay[aGameName];
        return aJoinName.playerJoin;
    }
    getGameInPlayHostPlayerName(aGameName){//gets the name of the host player from a game in play
        var aHostName = this.gamesInPlay[aGameName];
        return aHostName.playerHost;
    }
    getGameInPlayType(aGameName){//gets the gameType from a game in play
        var aGameType = this.gamesInPlay[aGameName];
        console.log(aGameType.gameType);
        return aGameType.gameType;
    }

    getGameInPlayData(aGameName){//handles retrieving the gameData.
        var aGameData = this.gamesInPlay[aGameName];
        //Line of code below from https://hackernoon.com/accessing-nested-objects-in-javascript-f02f1bd6387f
        const theData = ((aGameData || {}).gameData || {});//Good to have just in case, and works.
        return theData;
    }

    //Updates and data changes for both Games in Play and Games in Wait
    updateAndGetPlayerTurn(aGameName){//changes and returns the new player's turn
        console.log("updateAndGetPlayerTurn Called")
        var aTurn = this.gamesInPlay[aGameName];
        var changeTurn = !(aTurn.isHostTurn);
        this.gamesInPlay[aGameName].isHostTurn = changeTurn;
        return aTurn.isHostTurn;
    }

    updateGameData(newData, aGameName){//handles deleting and making a new gameData to update it.
        console.log("updateGameData Called");
        var oldData = this.getGameInPlayData(aGameName);
        delete this.gamesInPlay[aGameName].gameData;
        this.gamesInPlay[aGameName].gameData = newData;
        console.log("Changed? "+JSON.stringify(this.gamesInPlay[aGameName].gameData));
    }
    
    removeGameFromGIW(aGameKey){//returns removed key from games in wait
        console.log("removeGameFromGIW Called");
        var packData = {};
        packData[aGameKey] = this.getGameDataWithKey(aGameKey);
        var holdDataPack = Object.values(packData[aGameKey]);
        delete this.gamesInWait[aGameKey];
        return holdDataPack;
    }

    removeGameFromGIP(aGameKey){//returns removed key from games in wait
        console.log("removeGameFromGIP Called");
        delete this.gamesInPlay[aGameKey];
    }

    playerFoundGameTransfer(aGameKey, aPlayerJoining){//handles 'In wait' game, moves it to the 'In play' game list
        console.log("playerFoundGameTransfer Called");
        var holdData = this.removeGameFromGIW(aGameKey);
        var gameName = holdData[0];
        var gameType = holdData[1];
        var playerHost = holdData[2];
        var playerJoin = aPlayerJoining;
        var isHostTurn = true;
        var gameData = {};
        this.gamesInPlay[gameName] = {gameType, playerHost, playerJoin, isHostTurn, gameData};
    }
}