//Kiera Summersett
//This server is the main server, which handles communications between the rest of the server and the client-side code.

//==Permission Block, Essential Variables==
const express = require('express');
const path = require('path');
const bodyParser = require('body-parser');
const mongoose = require('mongoose');
const User = require('./model/user');
const bcrypt = require('bcrypt');
const jwt = require('jsonwebtoken');
const JWT_SECRET = 'Hd3rY4WgH04nTs5tGh0318N3G50tN';//PLACEHOLDER SECRET
const app = express();
const http = require("http");
const server = http.Server(app);
const socketio = require("socket.io");
const { group } = require('console');
const io = socketio(server);
mongoose.connect('mongodb://localhost:27017/login-app-db', function() {console.log("Mongo Connected.")});
app.use(express.static("pub"));
app.use(bodyParser.json());
const gamesInWait = require("./j_modules/GameManager");
const gameManager = new gamesInWait.GameManager();
const gamePlay = require("./j_modules/GamePlayer");
const gamePlayer = new gamePlay.GamePlayer();
let activePlayers = {};//Logged in Players are listed here

//==Functional==
function checkIfInWaitNames(nameToCheck){//This checks if the Game exists (is not yet fully in play)
    console.log("checkIfInWaitNames Called");
    const allNames = gameManager.getGameNames();
    for(i in allNames){
        if(nameToCheck.toLowerCase()===allNames[i].toLowerCase()) return true;
    }
    return false;
}
function getWaitingHostName(hostName){//This grabs the name of the game wanted using the host's username.
    console.log("getWaitingHostName Called");
    const allGameKeys = gameManager.getGameNames();
    for(i in allGameKeys){
        if(gameManager.getPlayerName(allGameKeys[i])===hostName) return allGameKeys[i];
    }
    return -1;
}
function getGameInPlayWithJoinName(joinName){//This grabs the name of the game in play wanted using the joining player's username
    console.log("getGameInPlayWithJoinName Called");
    const allGameKeys = gameManager.getGameInPlayNames();
    for(i in allGameKeys){
        if(gameManager.getGameInPlayJoinPlayerName(allGameKeys[i])===joinName) return allGameKeys[i];
    }
    return -1;
}
function getGameInPlayWithHostName(hostName){//This grabs the name of the game in play wanted using the host player's username
    console.log("getGameInPlayWithHostName Called");
    const allGameKeys = gameManager.getGameInPlayNames();
    for(i in allGameKeys){
        if(gameManager.getGameInPlayHostPlayerName(allGameKeys[i])===hostName) return allGameKeys[i];
    }
    return -1;
}
//==Account Handlers==
//Register, Login and changePassword code is mostly from https://www.youtube.com/watch?v=b91XgdyX-SM.
app.post('/api/register', async (req,res) => {//Handles new user registration
    console.log("register Called");
    const {username, password: plainTextPassword} = req.body;
    if(!username || typeof username !== 'string') {
        return res.json({status:'error', error: 'Invalid Username'});//ensures the name given is a string
    };
	if(!plainTextPassword || typeof plainTextPassword !== 'string') {
        return res.json({status:'error', error: 'Invalid Password'});//ensures the password given is a string
    };
    if(plainTextPassword.length <= 5) {
        return res.json({status:'error', error: 'Password must be at 6+ characters.'});
    };
    const password = await bcrypt.hash(plainTextPassword, 10);//encrypts the password, for security purposes
    try{
        const response = await User.create({username, password});
        console.log('User created successfuly: ', response);
    } catch (error) {
        if(error.code ===11000){//duplicate key
            return res.json({status: 'error', error: 'Username already in use'});
        };
        throw error;
    };
    res.json({status:'ok'});
});
app.post('/api/login', async (req, res) => {//Handles user login
    console.log("login Called");
    const {username, password} = req.body;
    const user = await User.findOne({username}).lean();
    if(!user){
        return res.json({status:'error',error: 'Invalid username/password'});
    }
    if(await bcrypt.compare(password, user.password)){//the username, password combination is successful
        const token = jwt.sign({ id: user._id, username: user.username }, JWT_SECRET);
        return res.json({ status:'ok',data: token });
    }
    res.json({status: 'error', data: 'Invalid username/password'});
});
app.post('/api/change-password', async (req,res)=>{//Handles changing the user's password
    console.log("change-password Called");
    const {newpassword: plainTextPassword, token} = req.body;
    if(!plainTextPassword || typeof plainTextPassword !== 'string') {
        return res.json({status:'error', error: 'Invalid Password'});
    };
    if(plainTextPassword.length < 5) {
        return res.json(
            {status:'error',error: 'Password should be at least 6 characters.'}
        );
    };
    try{
        const user = jwt.verify(token, JWT_SECRET);
        const _id = user.id;
        const password = await bcrypt.hash(plainTextPassword, 10);
        await User.updateOne({_id}, {$set: {password}});
        res.json({status:'ok'}); //console.log('JWT decoded: ', user);
    } catch (error){
        console.log(error);
        res.json({status: 'error', error:';))' });
    }
});

//==Game-Search Code==
app.post('/api/make-new-game', async (req, res)=>{//Handles Game-Search's game creation by a host user
    console.log("make-new-game Called");
    const {gameName, gameToPlay, playerName} = req.body;
    if(gameName&&gameToPlay&&playerName){
        const nameTaken = checkIfInWaitNames(gameName);
        if(nameTaken) return res.json({status: 'error', data: 'Name taken, choose another.'});
        if(gameName.length < 3) return res.json({status: 'error', data: 'The name must be three letters or more.'});
        gameManager.gamesInWait[gameName] = {gameToPlay, playerName};
        return res.json({status:'ok'});
    }else{
        return res.json({status: 'error', data: 'Fill out all required fields.'});
    }
});

/*----------------------------io Code-------------------------------*/
//==JWT authentication==
//from https://stackoverflow.com/questions/36788831/authenticating-socket-io-connections-using-jwt
io.use(function(socket, next){//Socket.io JWT verification, gives username back to page (initUsername)
    //io.use is run every time a connection is made.
    console.log("Verifying JWT...");
    if (socket.handshake.query && socket.handshake.query.token){
        jwt.verify(socket.handshake.query.token, JWT_SECRET, function(err, decoded) {
            if (err) return next(new Error('Authentication error'));
            socket.decoded = decoded;
            next();
            const namOfUser = decoded.username;
            let hasGame = false;
            let gameName = "";
            activePlayers[""+socket.id] = {namOfUser, hasGame, gameName};
            socket.on('initUsername', function(){ //for initializing username on webpages, which also runs when the page loads
                socket.emit('setUpUsername', namOfUser);
            });
        });
    }
    else {
        next(new Error('Authentication error'));
    }
});
//==Io Game Code==
io.on("connection", function(socket) {
	console.log("Somebody connected.");
    //==Game-Search Handlers==
    socket.on("getGamesForSearchList", function(){//Calls functions to retrieve gameManager's Games + their data, and sends it to the user.
        console.log("getGamesForSearchList Called");
        const gameNames = gameManager.getGameNames();
        const gameValues = gameManager.getGameValues();
        socket.emit("updatingGameSearchList",{gameNames, gameValues});
    });
    socket.on("getNewGameSearchList", function(){//Updates the current list of games like the function above, but updates all users.
        console.log("getNewGameSearchList Called");
        const gameNames = gameManager.getGameNames();
        const gameValues = gameManager.getGameValues();
        io.emit("updatingGameSearchList",{gameNames, gameValues});
    });
    socket.on("join-new-game", function(dataFromClient){//Handles joining another player (the host) game and setting the game to be "in play"
        console.log("join-new-game Called");
        const {gameKey, joinPlayerName} = dataFromClient;
        if(checkIfInWaitNames(gameKey)){
            gameManager.playerFoundGameTransfer(gameKey, joinPlayerName);//Will assure the player is found when they reach the game page
            socket.emit('joinGameResult', "OK");
        }else socket.emit('joinGameResult', "Sorry, game not found. Please try again.");
    });
    //==Game-Page==
    socket.on("handleHostFinding", function(dataFromClient){//checks if the game is in wait or play, and sends the appropriate message back.
        console.log("handleHostFinding Called");
        const playerName = dataFromClient;
        console.log(getWaitingHostName(playerName));
        console.log(socket.id);
        const gameIsCurrentlyWaiting = getWaitingHostName(playerName);
            if(gameIsCurrentlyWaiting!=-1){
                activePlayers[socket.id].hasGame = true;
                activePlayers[socket.id].gameName = gameIsCurrentlyWaiting;
                socket.emit("gameIsInWait", playerName);
            }
            else{
                //updates playerslist
                const tryGetWithAsHost = getGameInPlayWithHostName(playerName);
                const tryGetWithAsJoin = getGameInPlayWithJoinName(playerName);
                if(tryGetWithAsHost!=-1||tryGetWithAsJoin!=-1){//Check needed if error occurs between having a game in wait and a game in play.
                    socket.emit("gameIsInPlay", playerName);
                }
            }
    });
    socket.on("createHostGame", function(dataFromClient){//Handles Game-Page start, moves host to room
        console.log("createHostGame Called");
        const hostName = dataFromClient;
        const gameKey = getWaitingHostName(hostName);
        if(gameKey!=-1){
            const fullGameData = gameManager.getGameDataWithKey(gameKey);
            console.log(fullGameData);
            socket.join(gameKey);
            io.to(gameKey).emit('game-in-wait-for-2P', fullGameData);
        } else socket.emit('game-error-1', "Sorry, game not found. Please try again.");
    });
    socket.on("joinHostedGame", function(dataFromClient){//Handles 'hooking up' the player joining to the game itself, and calls the game setup.
        console.log("joinHostedGame Called");
        const joinerName = dataFromClient;
        const gameKey = getGameInPlayWithJoinName(joinerName);
        socket.join(gameKey);
        io.to(gameKey).emit('game-is-now-in-play', gameKey);
        startGame(gameKey);
    });
    
    //==GamePlay==
    function startGame(gameKey){//handles starting a new game, retrieving the proper data like the gameboard, and starting the playable game.
        console.log("startGame Called");
        const gameType = gameManager.getGameInPlayType(gameKey);
        const gameBoard = gamePlayer.initialGameMap(gameType);
        const playerTurn = gameManager.getGameInPlayHostPlayerName(gameKey);//PlayerTurn always starts with host turn being True
        io.to(gameKey).emit('game-ready', {gameBoard, playerTurn});
        //sends the new gameboard and the players turn. Always has host first, but is still here just in case.
    }
    socket.on("play-game", function(dataFromClient){//recieves data, sends it to the proper places to update the game, then sends it to the client.
        console.log("play-game Called");
        const {data, playerName, gameKey} = dataFromClient;
        const gameType = gameManager.getGameInPlayType(gameKey);
        const gameData = gameManager.getGameInPlayData(gameKey);
        const {gameBoard, aStatus, updateData} = gamePlayer.updateGameHandler(data, playerName, gameData, gameType);
        gameManager.updateGameData(updateData, gameKey);
        if(aStatus=="OKAY"){
            const updatedData = gameManager.getGameInPlayData(gameKey);//Updated data is only here as backup.
            const playerTurnIsHost = gameManager.updateAndGetPlayerTurn(gameKey);
            if(playerTurnIsHost){
                const playerTurn = gameManager.getGameInPlayHostPlayerName(gameKey);
                io.to(gameKey).emit('game-update', {gameBoard, playerTurn});
            }else{
                const playerTurn = gameManager.getGameInPlayJoinPlayerName(gameKey);
                io.to(gameKey).emit('game-update', {gameBoard, playerTurn});
            }
        }else{//A win state has been reached, and the game is done.
            const players = Object.keys(gameData);
            if(aStatus=="P1WIN"){//The Host Wins
                const winningPlayer = gameManager.getGameInPlayHostPlayerName(gameKey);
                const finalBoard = ("<p>"+winningPlayer
                +" has won the game!"+"</p>"
                +"<a href=http://localhost:6211/game-search.html>Go back to the Game Search?</a>");
                io.to(gameKey).emit('game-done', finalBoard);
            }else if(aStatus=="P2WIN"){//The Joining Player Wins
                const winningPlayer = gameManager.getGameInPlayJoinPlayerName(gameKey);
                const finalBoard = ("<p>"+winningPlayer
                +" has won the game!"+"</p>"
                +"<a href=http://localhost:6211/game-search.html>Go back to the Game Search?</a>");
                io.to(gameKey).emit('game-done', finalBoard);
            }else if(aStatus=="TIE"){//Occurs in games like Tic Tac Toe, where no one winning is a possible end.
                const winningPlayer = players[1];
                const finalBoard = ("<p>You both tied!</p>"
                +"<a href=http://localhost:6211/game-search.html>Go back to the Game Search?</a>");
                io.to(gameKey).emit('game-done', finalBoard);
            }
        }
    });
	socket.on("disconnect", function() {//Used if this socket connection was terminated (if the user went to a different page, etc.)
		console.log(socket.id+" disconnected. Finding Player...");
        console.log(Object.keys(activePlayers));
        console.log(Object.values(activePlayers));
        const playerIdList = Object.keys(activePlayers);
        for(i in playerIdList){//checks the entire list of active players
            if(socket.id==playerIdList[i]){
                if(activePlayers[socket.id].hasGame){//if the player found has a game, it finds and deletes it.
                    const inWait = checkIfInWaitNames(activePlayers[socket.id].gameName);
                    if(inWait){
                        gameManager.removeGameFromGIW(activePlayers[socket.id].gameName);
                    }else{
                        gameManager.removeGameFromGIP(activePlayers[socket.id].gameName);
                    }
                    //Updates the game-search so the nonexistant game is removed from the list.
                    const gameNames = gameManager.getGameNames();
                    const gameValues = gameManager.getGameValues();
                    io.emit("updatingGameSearchList",{gameNames, gameValues});
                }
                delete activePlayers[socket.id];
            }
        }
	});
});

server.listen(6211, function() {//A simple message stating the server is running.
	console.log("Server with socket.io is ready.");
});